{===============================================================================
 === Projet : GraphGen
 ===    Algorithe gntique pour rsoudre le problme du voyageur de commerce
 ===============================================================================
 === Auteur : Alexandre AUPETIT (aaupetit@club-internet.fr)
 ===    Voir mon site http://home.alex.tuxfamily.org/pvc.html
 ===     propos du problme du voyageur de commerce et des
 ===    algorithmes gntiques. Participez au dfi des 250 villes !
 ===    Vous trouverez galement mon programme tspgen, logiciel
 ===    libre crit en C++, et environ 50 fois plus performant
 ===    que celui-ci
 ===============================================================================
 === Licence : Logiciel libre sous licence GNU GPL
 ===    Ce logiciel doit toujours tre distribu avec ses sources
 ===    Vous avez le droit de modifier librement ces sources (et vous y
 ===    tes mme encourags)  condition que les modifications soient
 ===    galement libres (voir le fichier LICENCE.txt)
 ===============================================================================
 === Version  : 1.0.0
 === Date de cration : 10/08/1997
 === Date de dernire modif : 26/04/2003
 ===============================================================================
 === Fichier : unit d'affichage
 ===============================================================================
 === RAF :
 ==============================================================================}

unit uGraphGen;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, ComCtrls, StdCtrls, Buttons,
  Genutil, uPopulation, uAPropos;

type
  TfrmGraphique = class(TForm)
    ebNbVilles: TEdit;
    Label1: TLabel;
    UpDown1: TUpDown;
    pnlGraphique: TPanel;
    btCalcul: TBitBtn;
    Label2: TLabel;
    ebPourcInit: TEdit;
    Label3: TLabel;
    ebTempsCalcul: TEdit;
    Label7: TLabel;
    ebNbIndividus: TEdit;
    Label8: TLabel;
    ebNbSousPopulations: TEdit;
    btGenerer: TBitBtn;
    pbGraphique: TPaintBox;
    Label9: TLabel;
    ebNbMigr: TEdit;
    BitBtn1: TBitBtn;
    procedure FormResize(Sender: TObject);
    procedure btGenererClick(Sender: TObject);
    procedure btCalculClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormPaint(Sender: TObject);
    procedure BitBtn1Click(Sender: TObject);
  private
    { Dclarations prives }
  public
    { Dclarations publiques }
    iNbSousPopulations : integer;
    iPourcentageInit : integer;
    iNbMigrations : integer;
    iNbIndividus : integer;
    iTempsInit : integer;
    iTempsCalcul : integer;
    aptVille : array[1..500] of TPoint;
    iNbVilles : integer;
    Population : TPopulation;
    bAffichageEnCours : boolean;
    procedure AfficheVilles;
  end;

var
  frmGraphique: TfrmGraphique;

implementation

uses uIndividu;

{$R *.DFM}

function RecupereValeur(ebValeur : TEdit; iMin : integer; iMax : integer;
                                   iDefaut : integer; sNom : string) : integer;
begin
  // rcupration du nombre de SousPopulations
  try
    result := StrToInt(ebValeur.Text);
  except
    ShowMessage(sNom + ' doit tre entier...');
    ebValeur.SetFocus;
    result := iDefaut;
    ebValeur.Text := IntToStr(result);
  end;

  if result > iMax then
  begin
    ShowMessage(sNom + ' est trop grand...');
    result := iMax;
    ebValeur.Text := IntToStr(result);
    ebValeur.SetFocus;
   end;

  if result < iMin then
  begin
    ShowMessage(sNom + ' est trop petit...');
    result := iMin;
    ebValeur.Text := IntToStr(result);
    ebValeur.SetFocus;
  end;
end;


{===============================================================================
 === Fonction :
 ==============================================================================}
procedure TfrmGraphique.FormCreate(Sender: TObject);
begin
  btGenererClick(self);
  bAffichageEnCours := false;
end;


{===============================================================================
 === Fonction : Affiche les villes 
 ==============================================================================}
procedure TfrmGraphique.AfficheVilles;
var
  i : integer;
begin
  if not bAffichageEnCours then
  begin
    bAffichageEnCours := true;

    if Population = nil then
    begin
      //affichage des villes
      pbGraphique.Repaint;
      pbGraphique.Canvas.Brush.Style := bsSolid;
      pbGraphique.Canvas.Brush.Color := clRed;
      pbGraphique.Canvas.Pen.Color := clBlack;
      for i := 1 to iNbVilles do
        pbGraphique.Canvas.Rectangle(aptVille[i].X - 3, aptVille[i].Y - 3,aptVille[i].X + 3, aptVille[i].Y + 3);
    end
    else    
    begin
        Population.MeilleurIndividu.Affiche;
    end;
        
    bAffichageEnCours := false;
  end;
end;


{===============================================================================
 === Fonction : 
 ==============================================================================}
procedure TfrmGraphique.FormResize(Sender: TObject);
begin
  pnlGraphique.Top := 50;
  pnlGraphique.Height := frmGraphique.ClientHeight - 50;
  pnlGraphique.Left := 0;
  pnlGraphique.Width  := frmGraphique.ClientWidth;

  pbGraphique.Top := 0;
  pbGraphique.Height := pnlGraphique.Height;
  pbGraphique.Left := 0;
  pbGraphique.Width := pnlGraphique.Width;

  AfficheVilles;
end;




{===============================================================================
 === Fonction : Gnre les positions des villes
 ==============================================================================}
procedure TfrmGraphique.btGenererClick(Sender: TObject);
  var i : integer;
begin
  if Population <> nil then
  begin
      Population.Free;
      Population := nil;
  end;

  try
    iNbVilles := StrToInt(ebNbVilles.Text);

    // initialisation des villes alatoires
    for i := 1 to iNbVilles do
    begin
      aptVille[i].X := random(pbGraphique.Width-10) + 5;
      aptVille[i].Y := random(pbGraphique.Height-10) + 5;
    end;

    frmGraphique.Repaint;
  except
    ShowMessage('Le nombre de villes doit tre un entier...');
  end;
end;


{===============================================================================
 === Fonction : lance le calcul du plus court chemin entre les villes
 ==============================================================================}
procedure TfrmGraphique.btCalculClick(Sender: TObject);
var
  a , b : integer;
  AncienCurseur : TCursor;
begin
  try
  if Population <> nil then
      Population.Free;

    AncienCurseur := Screen.Cursor;
    Screen.Cursor := crHourGlass;
    iNbSousPopulations := 1;

    // rcupration du temps d'initialisation
    iPourcentageInit := RecupereValeur(ebPourcInit, 1,100, 10, 'Le pourcentage du temps d''initialisation');

    iNbMigrations := RecupereValeur(ebNbMigr, 1,1000, 5, 'Le nombre de migrations');

    iTempsCalcul := RecupereValeur(ebTempsCalcul, 1, 32700, 20, 'Le temps de calcul');

    iNbIndividus := RecupereValeur(ebNbIndividus, 1, NB_INDIVIDUS_MAX, 35, 'Le nombre d''individus');

    iNbSousPopulations := RecupereValeur(ebNbSousPopulations, 1, NB_SOUSPOPULATIONS_MAX, 1, 'Le nombre de sous-populations');

    pbGraphique.Repaint;

    Application.ProcessMessages;
    InitHasard;

    Population := TPopulation.Create(iNbSousPopulations, iNbMigrations, iPourcentageInit, iNbIndividus, teDarwin, tsPremiers,
                      iNbVilles, tmRenversement);

    Application.ProcessMessages;

    Population.Evolue(iTempsCalcul);

    MessageBeep(0);
    Application.ProcessMessages;

    finally
      Screen.Cursor := AncienCurseur;
    end;
end;

procedure TfrmGraphique.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  if Population <> nil then
      Population.Free;
end;

procedure TfrmGraphique.FormPaint(Sender: TObject);
begin
  AfficheVilles;
end;


procedure TfrmGraphique.BitBtn1Click(Sender: TObject);
begin
  frmAPropos.ShowModal;
end;

end.
