{===============================================================================
 === Projet : GraphGen
 ===    Algorithe gntique pour rsoudre le problme du voyageur de commerce
 ===============================================================================
 === Auteur : Alexandre AUPETIT (aaupetit@club-internet.fr)
 ===    Voir mon site http://home.alex.tuxfamily.org/pvc.html
 ===     propos du problme du voyageur de commerce et des
 ===    algorithmes gntiques. Participez au dfi des 250 villes !
 ===    Vous trouverez galement mon programme tspgen, logiciel
 ===    libre crit en C++, et environ 50 fois plus performant
 ===    que celui-ci
 ===============================================================================
 === Licence : Logiciel libre sous licence GNU GPL
 ===    Ce logiciel doit toujours tre distribu avec ses sources
 ===    Vous avez le droit de modifier librement ces sources (et vous y
 ===    tes mme encourags)  condition que les modifications soient
 ===    galement libres (voir le fichier LICENCE.txt)
 ===============================================================================
 === Version  : 1.0.0
 === Date de cration : 10/08/1997
 === Date de dernire modif : 26/04/2003
 ===============================================================================
 === Fichier : unit de la classe Individu
 ===============================================================================
 === RAF : Mutation
 ==============================================================================}

unit uIndividu;

interface

uses GenUtil;

type

TTabElement = array [1..1] of integer;
PTabElement = ^TTabElement ;


{===============================================================================
 === Classe : Individu : compos d'lments
 ==============================================================================}
TIndividu = class
  Element : PTabElement;
  hElement : Integer;
  NbElements : integer;
  Note : double;
  TypeMutation : TMut;
  constructor Create(NbElem : integer; TypeMut : TMut);
  procedure Initialise;
  procedure Evalue;
  procedure Mute;
  procedure AfficheEvaluation;
  procedure Affiche;
  procedure Efface;
  function TrouveElement(iNb : integer) : integer;
  destructor Destroy; override;
end;


implementation

uses Dialogs, uGraphGen, Graphics, SysUtils, Windows;



{==================================== * =======================================
 =============================== * INDIVIDU * =================================
 ==================================== * =======================================}


{===============================================================================
 === Fonction : Constructeur d'un individu
 ==============================================================================}
constructor TIndividu.Create(NbElem : integer; TypeMut : TMut);
   var i : integer;
begin
  Note := 0;
  NbElements := NbElem;
  TypeMutation := TypeMut;

  // rserve dynamiquement le tableau 
  hElement := GlobalAlloc(GMEM_FIXED, (NbElements+1) * sizeof(PTabElement));
  Element := GlobalLock(hElement);
  //... 
  //Element^[0]:= -1; 
 
  { on initialise l'individu avec les chiffres dans l'ordre }
  for i := 1 to NbElements do
    Element^[i] := i;
end;


{===============================================================================
 === Fonction :  Destructeur d'un individu
 ==============================================================================}
destructor TIndividu.Destroy;
   var i : integer;
begin
  // dsalloue le tableau dynamique
  GlobalUnlock(hElement);
  GlobalFree(hElement);
end;


{===============================================================================
 === Fonction : initialisation d'un individu
 ==============================================================================}
procedure TIndividu.Initialise;
   var a, b, i : integer;
begin
  { on melange en echangeant aleatoirement des elements }
  for i := 1 to NbElements do
  begin
    a := Hasard(NbElements);
    b := Hasard(NbElements);
    Echange(Element^[a], Element^[b]);
  end;

  { attention pas evalue }
end;


{===============================================================================
 === Fonction : Mutation d'un individu
 === - tmEchange : Erreur de replication : echange entre deux elements aleatoires
 === - tmEchangeRapproche : echanges rapproches : entre un element et le suivant
 ==============================================================================}
procedure TIndividu.Mute;
var
  a, b, i : integer;
begin
  {ShowMessage('Mutation');}
  case TypeMutation of
    tmEchange :
    begin
      a := Hasard(NbElements);
      b := Hasard(NbElements);
      Echange(Element^[a], Element^[b]);
    end;
    tmRenversement :
    begin
      a := Hasard(NbElements);
      b := Hasard(NbElements) div 2;
      for i := 1 to b do
        Echange(Element^[Circ(a-i, NbElements)],
                 Element^[Circ(a-1+i, NbElements)]);
    end;
    tmEchangeRapproche :
    begin
      a := Hasard(NbElements);
      b := ((Hasard(NbElements) + 1) mod NbElements) + 1;
      Echange(Element^[a], Element^[b]);
    end;
    tmDecalage :         { RAF }
    begin
      a := Hasard(NbElements);
      b := Hasard(NbElements);
      Echange(Element^[a], Element^[b]);
    end;
  end;
end;


{===============================================================================
 === Fonction : Evaluation d'un individu
 ==============================================================================}
procedure TIndividu.Evalue;
var
  i : integer;
begin
  Note := 0;
  for i := 1 to NbElements - 1 do
  begin
    Note := Note
      - abs(frmGraphique.aptVille[Element^[i]].X - frmGraphique.aptVille[Element^[i+1]].X)
      - abs(frmGraphique.aptVille[Element^[i]].Y - frmGraphique.aptVille[Element^[i+1]].Y);
  end;
  Note := Note
    - abs(frmGraphique.aptVille[Element^[NbElements - 1]].X - frmGraphique.aptVille[Element^[1]].X)
    - abs(frmGraphique.aptVille[Element^[NbElements - 1]].Y - frmGraphique.aptVille[Element^[1]].Y);

end;


{===============================================================================
 === Fonction :
 ==============================================================================}
procedure TIndividu.AfficheEvaluation;
begin
  //frmGraphique.txtNote.Caption := FloatToStr(Note);
end;


{===============================================================================
  === Fonction : affichage des composantes et de la note d'un individu
 ==============================================================================}
procedure TIndividu.Affiche;
var
  i : integer;
begin
  with frmGraphique do
  with frmGraphique.pbGraphique.Canvas do
  begin
    pbGraphique.Repaint;

    //affichage des villes
    Pen.Color := clBlack;
    Brush.Color := clRed;
    for i := 1 to NbElements do
      Rectangle(aptVille[i].X - 3, aptVille[i].Y - 3,aptVille[i].X + 3, aptVille[i].Y + 3);

    // affichager du parcours
    Pen.Color := clBlack;
    MoveTo(aptVille[Element^[1]].X, aptVille[Element^[1]].Y);
    for i := 2 to NbElements do
    begin
     LineTo(aptVille[Element^[i]].X, aptVille[Element^[i]].Y);
     MoveTo(aptVille[Element^[i]].X, aptVille[Element^[i]].Y);
    end;
    LineTo(aptVille[Element^[1]].X, aptVille[Element^[1]].Y);
  end;
end;


{===============================================================================
  === Fonction : effacement  l'cran d'un individu
 ==============================================================================}
procedure TIndividu.Efface;
var
  i : integer;
begin
  with frmGraphique do
  with frmGraphique.pbGraphique.Canvas do
  begin
    // effacement du parcours
    Pen.Color := clWindow;
    MoveTo(aptVille[Element^[1]].X, aptVille[Element^[1]].Y);
    for i := 2 to NbElements do
    begin
     LineTo(aptVille[Element^[i]].X, aptVille[Element^[i]].Y);
     MoveTo(aptVille[Element^[i]].X, aptVille[Element^[i]].Y);
    end;
    LineTo(aptVille[Element^[1]].X, aptVille[Element^[1]].Y);
  end;  
end;



{========================================================================
  === Fonction : determine j tel que I[j] = nb (renvoie 0 sinon)
  === Entree : dim : dimension de l'individu
  ===          I : individu
  ===          nb : l'entier a touver
  === Retour : j
  ========================================================================}
function TIndividu.TrouveElement(iNb : integer) : integer;
var
  iCompt : integer;
begin
  result := 0;
  iCompt := 1;
  repeat
    if Element^[iCompt] = iNb then
      result := iCompt;
    Inc(iCompt);
  until ((result <> 0) or (iCompt > NbElements))
end;


end.
