{===============================================================================
 === Projet : GraphGen
 ===    Algorithe gntique pour rsoudre le problme du voyageur de commerce
 ===============================================================================
 === Auteur : Alexandre AUPETIT (aaupetit@club-internet.fr)
 ===    Voir mon site http://home.alex.tuxfamily.org/pvc.html
 ===     propos du problme du voyageur de commerce et des
 ===    algorithmes gntiques. Participez au dfi des 250 villes !
 ===    Vous trouverez galement mon programme tspgen, logiciel
 ===    libre crit en C++, et environ 50 fois plus performant
 ===    que celui-ci
 ===============================================================================
 === Licence : Logiciel libre sous licence GNU GPL
 ===    Ce logiciel doit toujours tre distribu avec ses sources
 ===    Vous avez le droit de modifier librement ces sources (et vous y
 ===    tes mme encourags)  condition que les modifications soient
 ===    galement libres (voir le fichier LICENCE.txt)
 ===============================================================================
 === Version  : 1.0.0
 === Date de cration : 10/08/1997
 === Date de dernire modif : 26/04/2003
 ===============================================================================
 === Fichier : unit de la classe Population
 ===============================================================================
 === RAF : Liste au lieu de tableaux
 ==============================================================================}

unit uPopulation;

interface

uses Genutil, uSousPopulation, uIndividu;

const
  NB_SOUSPOPULATIONS_MAX = 100;

type
  TTabSousPopulation = array[1..NB_SOUSPOPULATIONS_MAX] of TSousPopulation;

{===============================================================================
 === Classe :Population : Ensemble d'Individus
 ===============================================================================}
TPopulation = class
  SousPopulation : TTabSousPopulation;
  NbSousPopulations : integer;
  NbMigrations : integer;
  NbIndividus : integer;
  PourcentageInitialisation : real;
  TypeEvolution : TEvol;
  TypeSelection : TSel;
  NbElements : integer;
  TypeMutation : TMut;
  MeilleurIndividu : TIndividu;
  constructor Create(NbSousPop :integer; NbMigr : integer; PourcInit : real; NbIndiv : integer; TypeEvol : TEvol; TypeSel : TSel;
              NbElem : integer;
              TypeMut : TMut);
  procedure Evolue(iTempsTotal : integer);
  destructor Destroy; override;
end;

implementation

uses Windows, Dialogs, Sysutils;

{==================================== * =======================================
 ============================== * POPULATION * ================================
 ==================================== * =======================================}


{===============================================================================
 === Fonction :  Constructeur d'une Population
 ==============================================================================}
constructor TPopulation.Create(nbSousPop : integer; NbMigr: integer; PourcInit : real; NbIndiv : integer; TypeEvol : TEvol; TypeSel : TSel;
            NbElem : integer; TypeMut : TMut);
var
 i : integer;
begin
  inherited Create;

  NbSousPopulations := NbSousPop;
  NbMigrations := NbMigr;
  PourcentageInitialisation := PourcInit;
  NbIndividus := NbIndiv;
  TypeEvolution := TypeEvol;
  TypeSelection := TypeSel;
  NbElements := NbElem;
  TypeMutation := TypeMut;

  for i := 1 to NbSousPopulations do       (*********)
    SousPopulation[i] := TSousPopulation.Create(NbIndividus, TypeEvolution, TypeSelection,
                      NbElements, TypeMutation);

  MeilleurIndividu := TIndividu.Create(NbElem, TypeMut);
end;


{===============================================================================
 === Fonction :  Destructeur d'une Population
 ==============================================================================}
destructor TPopulation.Destroy;
var
  i : integer;
begin
  for i := 1 to NbSousPopulations do
        SousPopulation[i].Free;

  MeilleurIndividu.Free;

  inherited Destroy;
end;


{===============================================================================
 === Fonction : fait voluer une Population
 ==============================================================================}
procedure TPopulation.Evolue(iTempsTotal : integer);
var
  i, a, b, iCompt, iTempsEvol, iTempsInit : integer;
begin
    iTempsInit := round(iTempsTotal * PourcentageInitialisation / (100*NbSousPopulations));
    //ShowMessage(' temps init : ' + IntToStr(iTempsInit));
    // optimisation  du meilleur de chaque SousPopulation
    for i := 1 to NbSousPopulations do
    begin
      SousPopulation[i].EvolueHasard(iTempsInit);
    end;

    iTempsEvol := round( (iTempsTotal - (iTempsTotal * PourcentageInitialisation / 100)) / (NbMigrations*NbSousPopulations) );
    //ShowMessage(' temps vol : ' + IntToStr(iTempsEvol));
    for i := 1 to NbMigrations do
    begin
      // volution de chaque SousPopulation
      for i := 1 to NbSousPopulations do
      begin
        SousPopulation[i].Evolue(iTempsEvol);
      end;

      // migration des individus entre les SousPopulations
      if NbSousPopulations > 1 then
      begin
        for i := 1 to round(NbSousPopulations * NbIndividus / 5) do
        begin
          a := Hasard(NbSousPopulations);
          repeat
            b := Hasard(NbSousPopulations);
          until (b <> a);
          SousPopulation[a].Reinsere(SousPopulation[b].Individu[Hasardx3(NbIndividus)]);
        end;
      end;
    end;

    for iCompt := 1 to NbElements do
      MeilleurIndividu.Element^[iCompt] := SousPopulation[1].Individu[1].Element^[iCompt];
    MeilleurIndividu.Note :=  SousPopulation[1].Individu[1].Note;

    if NbSousPopulations > 1 then
    begin
      for i := 2 to NbSousPopulations do
      begin
        if SousPopulation[i].Individu[1].Note > MeilleurIndividu.Note then
        begin
          for iCompt := 1 to NbElements do
            MeilleurIndividu.Element^[iCompt] := SousPopulation[1].Individu[1].Element^[iCompt];
          MeilleurIndividu.Note :=  SousPopulation[1].Individu[1].Note;
//          MeilleurIndividu.Affiche;
        end;
      end;
    end;

    MessageBeep(0);
//    MeilleurIndividu.Affiche;
end;

end.

