{===============================================================================
 === Projet : GraphGen
 ===    Algorithe gntique pour rsoudre le problme du voyageur de commerce
 ===============================================================================
 === Auteur : Alexandre AUPETIT (aaupetit@club-internet.fr)
 ===    Voir mon site http://home.alex.tuxfamily.org/pvc.html
 ===     propos du problme du voyageur de commerce et des
 ===    algorithmes gntiques. Participez au dfi des 250 villes !
 ===    Vous trouverez galement mon programme tspgen, logiciel
 ===    libre crit en C++, et environ 50 fois plus performant
 ===    que celui-ci
 ===============================================================================
 === Licence : Logiciel libre sous licence GNU GPL
 ===    Ce logiciel doit toujours tre distribu avec ses sources
 ===    Vous avez le droit de modifier librement ces sources (et vous y
 ===    tes mme encourags)  condition que les modifications soient
 ===    galement libres (voir le fichier LICENCE.txt)
 ===============================================================================
 === Version  : 1.0.0
 === Date de cration : 10/08/1997
 === Date de dernire modif : 26/04/2003
 ===============================================================================
 === Fichier : unit de la classe Sous-Population (groupe)
 ===============================================================================
 === RAF : Liste au lieu de tableaux
 ==============================================================================}

unit uSousPopulation;

interface

uses uIndividu, GenUtil;

type

TTabIndividu = array[1..NB_INDIVIDUS_MAX] of TIndividu;

{===============================================================================
 === Classe : SousPopulation : Ensemble d'Individus
 ===============================================================================}
TSousPopulation = class
  Individu : TTabIndividu;
  NbIndividus : integer;
  TypeEvolution : TEvol;
  TypeSelection : TSel;
  NbElements : integer;
  TypeMutation : TMut;
  constructor Create(NbIndiv : integer; TypeEvol : TEvol; TypeSel : TSel;
              NbElem : integer; TypeMut : TMut);
  procedure Classe;
  procedure Evolue(iTemps : integer);
  procedure EvolueHasard(iTemps : integer);
  procedure EvolueMonteCarlo(iTemps : integer);
  procedure EvolueDarwin(iTemps : integer);
  procedure EchangeIndividus(i : integer; j : integer);
  procedure Selectionne(var a : integer; var b : integer);
  function Appartient(I : TIndividu) : boolean;
  function Reinsere(I : TIndividu) : integer;
  procedure Recombine(a : integer;  b : integer;  F : TIndividu);
  destructor Destroy; override;
end;


implementation

uses SysUtils, Dialogs, Windows;



{==================================== * =======================================
 =========================== * SOUS - POPULATION * ============================
 ==================================== * =======================================}


{===============================================================================
 === Fonction :  Constructeur d'une SousPopulation
 ==============================================================================}
constructor TSousPopulation.Create(NbIndiv : integer; TypeEvol : TEvol; TypeSel : TSel;
            NbElem : integer; TypeMut : TMut);
var
 i : integer;
begin
  inherited Create;
  NbIndividus := NbIndiv;
  TypeEvolution := TypeEvol;
  TypeSelection := TypeSel;
  NbElements := NbElem;
  TypeMutation := TypeMut;

  { cree les individus }
  for i := 1 to NbIndividus do
  begin
    Individu[i] := TIndividu.Create(NbElem, TypeMut);
    {on l'initialise}
    Individu[i].Initialise;
    { on evalue sa note }
    Individu[i].Evalue;
  end;

  {les classe dans l'ordre}
  Classe;
end;


{===============================================================================
 === Fonction :  Destructeur d'une SousPopulation
 ==============================================================================}
destructor TSousPopulation.Destroy;
var
  i : integer;
begin
  for i := 1 to NbIndividus do
    Individu[i].Free;

  inherited Destroy;
end;


(* ASSIGN ?*)
// inter = membre de la sspop ???? (pa  rinitialiser  chaque fois...)
{===============================================================================
 === Fonction : Echange les individus d'indice i et j
 ==============================================================================}
procedure TSousPopulation.EchangeIndividus(i : integer; j : integer);
var
  iCompt : integer;
  Inter : PTabElement;
  hInter : integer;
  InterNote : real;
begin
  // rserve dynamiquement le tableau 
  hInter := GlobalAlloc(GMEM_FIXED, (NbElements+1) * sizeof(PTabElement));
  Inter := GlobalLock(hInter);

  for iCompt := 1 to NbElements do
  begin
    Inter^[iCompt] := Individu[i].Element^[iCompt];
    Individu[i].Element^[iCompt] := Individu[j].Element^[iCompt];
    Individu[j].Element^[iCompt] := Inter^[iCompt];
  end;
  InterNote := Individu[i].Note;
  Individu[i].Note := Individu[j].Note;
  Individu[j].Note := InterNote;

  // dsalloue le tableau dynamique
  GlobalUnlock(hInter);
  GlobalFree(hInter);
end;


{===============================================================================
 === Fonction : Tri a bulles - classe les individus de la SousPopulation par
 ===            note decroissante
 ==============================================================================}
procedure TSousPopulation.Classe;
var
  i, j : integer;
begin
  for i := 2 to NbIndividus do
    for j := 1 to i - 1 do
      if Individu[i].Note > Individu[j].Note then
        EchangeIndividus(i, j);
end;


{===============================================================================
 === Fonction : Slection des deux individus parents pour la recombinaison
 ==============================================================================}
procedure TSousPopulation.Selectionne(var a : integer; var b : integer);
begin
  case TypeSelection of
    tsHasard :
    begin
      a := Hasard(NbIndividus);
      b := Hasard(NbIndividus);
    end;
    tsPremiers :
    begin
      {a := Hasardx2(NbIndividus);
      b := Hasardx2(NbIndividus);}
      a := Hasardx3(NbIndividus);   (* + roulette wheel ?*)
      b := Hasardx3(NbIndividus);
    end;
    tsProche :
    begin
      a := Hasardx2(NbIndividus);
      b := ((Hasard(NbIndividus) + 1) mod NbIndividus) + 1;
    end;
  end;
end;



{===============================================================================
 === Fonction :  recombinaison des deux individus parents pour la
 === - Crossover a 1 point
 ==============================================================================}
procedure TSousPopulation.Recombine(a : integer;  b : integer;  F : TIndividu);
var
  iCompt1, iCompt2, iChange : integer;
begin
  iChange := Hasard(NbElements);

  { jusqu'a iChange on recopie le parent 1 dans le fils }
  for iCompt1 := 1 to iChange - 1 do
    F.Element^[iCompt1] := Individu[a].Element^[iCompt1];

  { ensuite on remplit de 0 }
  for iCompt1 := iChange to NbElements do
    F.Element^[iCompt1] := 0;

  { on trouve l'index de element iChange dans I2
  iCompt2 = TrouveElement(dim, I2, I1[iChange]); }

  iCompt2 := iChange;

  for iCompt1 := iChange to NbElements do
  begin                                                  { tableau invers ? pour trouveelement}
    { On recherche dans le parent 2 la premiere ville
    non encore visitee par le fils }
    while F.TrouveElement(Individu[b].Element^[iCompt2]) > 0 do
      iCompt2 := (iCompt2 mod NbElements) + 1; { ?????????????}

    F.Element^[iCompt1] := Individu[b].Element^[iCompt2];
  end;
end;



{===============================================================================
 === Fonction : recherche si un individu appartient deja a une SousPopulation
 === Entree : I : l'individu a tester
 === Retour : true si l'individu appartient a la SousPopulation,
 ===          false sinon
 ==============================================================================}
function TSousPopulation.Appartient(I : TIndividu) : boolean;
var
  iCompt1, iCompt2 : integer;
begin
  result := false;
  iCompt2 := 10;
  { on compare l'individu I a chaque individu de la SousPopulation P1 }
  while (iCompt2 <= NbIndividus) and (not result) do
  begin
    { si les notes sont les memes  (sinon les individus sont differents) }
    if (I.Note = Individu[iCompt2].Note) then
    begin
      result := true;
      iCompt1 := 1;
      { on regarde pour chaque element de l'individu ... }
      while (result) and (iCompt1 <= NbElements) do
      begin
        { ... si un element est different }
        if (I.Element^[iCompt1] <> Individu[iCompt2].Element^[iCompt1]) then
           result := false; { des qu'au moins un est different ont sort }
                            { de la boucle et on renvoie faux }
        Inc(iCompt1);
      end;
      { si tous les elements sont egaux on renvoie vrai }
    end;
    Inc(iCompt2);
  end;
end;

{recherche par dichotomie - enlever les cas extrmes...}
{ Fonctions : TrouvePlace ... combiner avec appartient
Insere(I, place) :

function TSousPopulation.Reinsere(I : TIndividu) : integer;
begin
  result := TrouvePlace(I);
  if result <> 0 then
    Insere(I, result);
end;
}

(*
function TSousPopulation.Reinsere2(I : TIndividu) : integer;
var iMin, iMax, i, j : integer;
begin
  {si la note de l'individu est plus petite que le mons bon lment de la SousPopulation }
  if I.Note < Individu[iMax].Note then
    result := 0
  {si la note de l'individu est plus grande que le meilleur lment de la SousPopulation }
  else if I.Note > Individu[iMin].Note then
    result := 1
  { si entre les deux }
  else
  begin
    { recherche de la note par dichotomie }
    result := 0;
    iMin := 1;
    iMax := NbIndividus;
    repeat
      if (I.Note = Individu[iMin].Note) then
        result := iMin
      else if I.Note = Individu[iMax].Note then
        result := iMax
      { si intervalle = 1 => entre les deux ????????????????? }
      else if (iMin = iMax - 1)
        result := iMax
      else
      begin
        iMilieu := (iMin + iMax) div 2;
        if I.Note > Individu[iMilieu].Note
          iMin := iMilieu
        else
          iMax := iMilieu;
      end;
    until result <> 0;

    { si l'individu appartient dj  la SousPopulation }
    if Egal(I, result) then
      { on ne le prend pas en compte }
      result := 0;
  end;

  { si le nouvel individu a trouv sa place dans la SousPopulation }
  if result <> 0 then
  begin
    { on l'insre en dcalant les autres }
    { dcale les individus placs aprs }
    for j := NbIndividus downto result + 1 do
    begin
      for i := 1 to NbElements do
        Individu[j].Element^[i] := Individu[j - 1].Element^[i];
      Individu[j].Note := Individu[j - 1].Note;
    end;
    { insre le nouvel lment }
    for i := 1 to NbElements do
      Individu[result].Element^[i] := I.Element^[i];
    Individu[result].Note := I.Note;
  end;
end;         *)



{===============================================================================
 === Fonction : Tri : classe un individu d'une SousPopulation deja triee
 ===            par note decroissante
 === Entree : I : l'individu a inserer
 === Retour : la place qu'a prise l'individu dans le groupe, -1 sinon
 ==============================================================================}
function TSousPopulation.Reinsere(I : TIndividu) : integer;
var
  iCompt1, iCompt2, j : integer;
  bOk : boolean;
begin
  iCompt2 := -1;
  { Si la note de l'individu cree est meilleure que celle du groupe }
  if I.Note > Individu[NbIndividus].Note then
  begin
    { si l'individu n'appartient pas deja a la SousPopulation }
    if not Appartient(I) then
      begin
        iCompt2 := 1;
        bOk := false;
        repeat
          if I.Note > Individu[iCompt2].Note then
          begin
            bOk := true;
            { dcale les individus placs aprs }
            for j := NbIndividus downto iCompt2 + 1 do
            begin
              for iCompt1 := 1 to NbElements do
                Individu[j].Element^[iCompt1] := Individu[j - 1].Element^[iCompt1];
              Individu[j].Note := Individu[j - 1].Note;
            end;
            { insre le nouvel lment }
            for iCompt1 := 1 to NbElements do
              Individu[iCompt2].Element^[iCompt1] := I.Element^[iCompt1];
            Individu[iCompt2].Note := I.Note;
          end
          else
            iCompt2 := iCompt2 + 1;
        until (bOk) or (iCompt2 > NbIndividus);
      end;
  end;
  result := iCompt2;
end;


{===============================================================================
 === Fonction : cree des individus au hasard pendant un temps donne
 ===            les NbIndividus meilleurs sont ranges dans la SousPopulation
 ==============================================================================}
procedure TSousPopulation.EvolueHasard(iTemps : integer);
var
 { iParent1, iParent2 : integer; }
 I : TIndividu;             { individu aleatoire }
 iPlace : integer;          { la place obtenue par cet individu dans la SousPopulation }
 iNbSec : integer;          { le nombre de secondes ecoulees depuis le debut du calcul }
 iTempo : integer;          { un compteur }
 dtTempsDebut : TDateTime;  { le temps de debut de calcul }
 ch : char;
 iCompt : integer;
begin
  { Selectionne(iParent1, iParent2);   }

  dtTempsDebut := Time;
  iNbSec := 0;
  iTempo := 0;

  { Cre un individu }
  I := TIndividu.Create(NbElements, TypeMutation); { cr 1 fois puis rinitialis}
  repeat
    { Initialise et evalue un individu au hasard }
    I.Initialise;
    I.Evalue;
    Inc(iTempo);
    { pour eviter de calculer le temps passe a chaque fois que l'on
      cree un individu, on utilise le compteur iTempo :
      toutes les TEMPO_MAX iterations, on calcule le temps passe }
    if iTempo > TEMPO_MAX  then
    begin
      iTempo := 0;
      { on calcule la difference de temps en secondes
        entre le debut du calcul (dtTempsDebut) et maintenant }
      iNbSec := round((Time - dtTempsDebut) *  86400);
      (*tac;*)
    end;
    { REINSERTION : si la note de l'individu I est plus grande que la moins
      bonne note des individus de la SousPopulation, on l'insere
      dans cette SousPopulation }
    iPlace := Reinsere(I);
    { si l'individu a pris la premiere place }
    if iPlace = 1 then
    begin
      { on affiche le nouveau meilleur score }
      Individu[1].Affiche;
    end;
  until (iNbSec > iTemps);
  { on libre l'individu courant }
  I.Free;
  { on affiche la meilleure note de la SousPopulation }
  Individu[1].Affiche;
  { et le temps effectif de calcul  }
  //writeln('Duree du calcul : ', iNbSec, ' secondes');
end;


{===============================================================================
 === Fonction : cree des individus par mutation pendant un temps donne
 ===            les meilleurs sont ranges dans la SousPopulation
 ==============================================================================}
procedure TSousPopulation.EvolueMonteCarlo(iTemps : integer);
var
 { iParent1, iParent2 : integer; }
 I : TIndividu;             { individu aleatoire }
 iParent : integer;         { index du parent }
 iPlace : integer;          { la place obtenue par cet individu dans la SousPopulation }
 iNbSec : integer;          { le nombre de secondes ecoulees depuis le debut du calcul }
 iTempo : integer;          { un compteur }
 iCompt1 : integer;          { un compteur }
 dtTempsDebut : TDateTime;  { le temps de debut de calcul }
 ch : char;
 iCompt : integer;
begin
  dtTempsDebut := Time;
  iNbSec := 0;
  iTempo := 0;
  
  { Cre un individu }
  I := TIndividu.Create(NbElements, TypeMutation); { cr 1 fois puis rinitialis}
  repeat
    {Cree un nouvel individu en mutant un individu de la SousPopulation }
    iParent := Hasardx2(NbIndividus);
    for iCompt1 := 1 to NbElements do
      I.Element^[iCompt1] := Individu[iParent].Element^[iCompt1];
    I.Mute;
    I.Evalue;
    Inc(iTempo);
    { pour eviter de calculer le temps passe a chaque fois que l'on
      cree un individu, on utilise le compteur iTempo :
      toutes les TEMPO_MAX iterations, on calcule le temps passe }
    if iTempo > TEMPO_MAX  then
    begin
      iTempo := 0;
      { on calcule la difference de temps en secondes
        entre le debut du calcul (dtTempsDebut) et maintenant }
      iNbSec := round((Time - dtTempsDebut) *  86400);
      (*tac; *)
    end;
    { REINSERTION : si la note de l'individu I est plus grande que la moins
      bonne note des individus de la SousPopulation, on l'insere
      dans cette SousPopulation }
    iPlace := Reinsere(I);
    { si l'individu a pris la premiere place }
    if iPlace = 1 then
    begin
      { on affiche le nouveau meilleur score }
      Individu[1].Affiche;
    end;
  until (iNbSec > iTemps);
  { on libre l'individu courant }
  I.Free;
  { on affiche la meilleure note de la SousPopulation }
  Individu[1].Affiche;
  { et le temps effectif de calcul  }
  //writeln('Duree du calcul : ', iNbSec, ' secondes');
end;


{===============================================================================
 === Fonction : cree des individus par recombinaison et mutation pendant un
 ===            temps donne. Les meilleurs sont ranges dans la SousPopulation
 ==============================================================================}
procedure TSousPopulation.EvolueDarwin(iTemps : integer);
var
 { iParent1, iParent2 : integer; }
 I : TIndividu;                { individu aleatoire }
 iParent1, iParent2 : integer; { index des parents }
 iPlace : integer;             { la place obtenue par cet individu dans la SousPopulation }
 iNbSec : integer;             { le nombre de secondes ecoulees depuis le debut du calcul }
 iTempo : integer;             { un compteur }
 iCompt1 : integer;            { un compteur }
 dtTempsDebut : TDateTime;     { le temps de debut de calcul }
 ch : char;
 iCompt : integer;
begin
  dtTempsDebut := Time;
  iNbSec := 0;
  iTempo := 0;

  { Cre un individu }
  I := TIndividu.Create(NbElements, TypeMutation); { cr 1 fois puis rinitialis}
  repeat
    {Cree un nouvel individu par recombinaison de deux individus de la SousPopulation }
    Selectionne(iParent1, iParent2);
    Recombine(iParent1, iParent2, I); {********* fonction??? ***}
    I.Mute;
    I.Evalue;
    Inc(iTempo);
    { pour eviter de calculer le temps passe a chaque fois que l'on
      cree un individu, on utilise le compteur iTempo :
      toutes les TEMPO_MAX iterations, on calcule le temps passe }
    if iTempo > TEMPO_MAX  then
    begin
      iTempo := 0;
      { on calcule la difference de temps en secondes
        entre le debut du calcul (dtTempsDebut) et maintenant }
      iNbSec := round((Time - dtTempsDebut) *  86400);
      (*tac;*)
    end;
    { REINSERTION : si la note de l'individu I est plus grande que la moins
      bonne note des individus de la SousPopulation, on l'insere
      dans cette SousPopulation }
    iPlace := Reinsere(I);
    { si l'individu a pris la premiere place }
    if iPlace = 1 then
    begin
      { on affiche le nouveau meilleur score }
      Individu[1].Affiche;
    end;
  until (iNbSec > iTemps);
  { on libre l'individu courant }
  I.Free;
  { on affiche la meilleure note de la SousPopulation }
  Individu[1].Affiche;
  { et le temps effectif de calcul  }
  //writeln('Duree du calcul : ', iNbSec, ' secondes');
end;


{===============================================================================
 === Fonction : fait voluer une SousPopulation
 ==============================================================================}
procedure TSousPopulation.Evolue(iTemps : integer);
begin
  case TypeEvolution of
  teMonteCarlo :
    EvolueMonteCarlo(iTemps);
  teDarwin :
    EvolueDarwin(iTemps);
  end;
end;



end.
